#include "common/types/types.h"

#include <stdexcept>

#include "common/exception.h"
#include "common/types/types_include.h"

namespace kuzu {
namespace common {

DataType::DataType() : typeID{ANY}, childType{nullptr} {}

DataType::DataType(DataTypeID typeID) : typeID{typeID}, childType{nullptr} {
    assert(typeID != LIST);
}

DataType::DataType(DataTypeID typeID, std::unique_ptr<DataType> childType)
    : typeID{typeID}, childType{std::move(childType)} {
    assert(typeID == LIST);
}

DataType::DataType(const DataType& other) : typeID{other.typeID} {
    if (other.childType) {
        childType = other.childType->copy();
    }
}

DataType::DataType(DataType&& other) noexcept
    : typeID{other.typeID}, childType{std::move(other.childType)} {}

std::vector<DataTypeID> DataType::getNumericalTypeIDs() {
    return std::vector<DataTypeID>{INT64, DOUBLE};
}

std::vector<DataTypeID> DataType::getAllValidTypeIDs() {
    return std::vector<DataTypeID>{
        INTERNAL_ID, BOOL, INT64, DOUBLE, STRING, DATE, TIMESTAMP, INTERVAL, LIST};
}

DataType& DataType::operator=(const DataType& other) {
    typeID = other.typeID;
    if (other.childType) {
        childType = other.childType->copy();
    }
    return *this;
}

bool DataType::operator==(const DataType& other) const {
    if (typeID != other.typeID) {
        return false;
    }
    if (typeID == LIST && *childType != *other.childType) {
        return false;
    }
    return true;
}

bool DataType::operator!=(const DataType& other) const {
    return !((*this) == other);
}

DataType& DataType::operator=(DataType&& other) noexcept {
    typeID = other.typeID;
    childType = std::move(other.childType);
    return *this;
}

std::unique_ptr<DataType> DataType::copy() {
    if (childType) {
        return make_unique<DataType>(typeID, childType->copy());
    } else {
        return std::make_unique<DataType>(typeID);
    }
}

DataTypeID DataType::getTypeID() const {
    return typeID;
}

DataType* DataType::getChildType() const {
    return childType.get();
}

DataType Types::dataTypeFromString(const std::string& dataTypeString) {
    DataType dataType;
    if (dataTypeString.ends_with("[]")) {
        dataType.typeID = LIST;
        dataType.childType = std::make_unique<DataType>(
            dataTypeFromString(dataTypeString.substr(0, dataTypeString.size() - 2)));
        return dataType;
    } else {
        dataType.typeID = dataTypeIDFromString(dataTypeString);
    }
    return dataType;
}

DataTypeID Types::dataTypeIDFromString(const std::string& dataTypeIDString) {
    if ("INTERNAL_ID" == dataTypeIDString) {
        return INTERNAL_ID;
    } else if ("INT64" == dataTypeIDString) {
        return INT64;
    } else if ("DOUBLE" == dataTypeIDString) {
        return DOUBLE;
    } else if ("BOOLEAN" == dataTypeIDString) {
        return BOOL;
    } else if ("STRING" == dataTypeIDString) {
        return STRING;
    } else if ("DATE" == dataTypeIDString) {
        return DATE;
    } else if ("TIMESTAMP" == dataTypeIDString) {
        return TIMESTAMP;
    } else if ("INTERVAL" == dataTypeIDString) {
        return INTERVAL;
    } else {
        throw Exception("Cannot parse dataTypeID: " + dataTypeIDString);
    }
}

std::string Types::dataTypeToString(const DataType& dataType) {
    if (dataType.typeID == LIST) {
        assert(dataType.childType);
        auto result = dataTypeToString(*dataType.childType) + "[]";
        return result;
    } else {
        return dataTypeToString(dataType.typeID);
    }
}

std::string Types::dataTypeToString(DataTypeID dataTypeID) {
    switch (dataTypeID) {
    case ANY:
        return "ANY";
    case NODE:
        return "NODE";
    case REL:
        return "REL";
    case INTERNAL_ID:
        return "INTERNAL_ID";
    case BOOL:
        return "BOOL";
    case INT64:
        return "INT64";
    case DOUBLE:
        return "DOUBLE";
    case DATE:
        return "DATE";
    case TIMESTAMP:
        return "TIMESTAMP";
    case INTERVAL:
        return "INTERVAL";
    case STRING:
        return "STRING";
    case LIST:
        return "LIST";
    default:
        assert(false);
    }
}

std::string Types::dataTypesToString(const std::vector<DataType>& dataTypes) {
    std::vector<DataTypeID> dataTypeIDs;
    for (auto& dataType : dataTypes) {
        dataTypeIDs.push_back(dataType.typeID);
    }
    return dataTypesToString(dataTypeIDs);
}

std::string Types::dataTypesToString(const std::vector<DataTypeID>& dataTypeIDs) {
    if (dataTypeIDs.empty()) {
        return {""};
    }
    std::string result = "(" + Types::dataTypeToString(dataTypeIDs[0]);
    for (auto i = 1u; i < dataTypeIDs.size(); ++i) {
        result += "," + Types::dataTypeToString(dataTypeIDs[i]);
    }
    result += ")";
    return result;
}

uint32_t Types::getDataTypeSize(DataTypeID dataTypeID) {
    switch (dataTypeID) {
    case INTERNAL_ID:
        return sizeof(internalID_t);
    case BOOL:
        return sizeof(uint8_t);
    case INT64:
        return sizeof(int64_t);
    case DOUBLE:
        return sizeof(double_t);
    case DATE:
        return sizeof(date_t);
    case TIMESTAMP:
        return sizeof(timestamp_t);
    case INTERVAL:
        return sizeof(interval_t);
    case STRING:
        return sizeof(ku_string_t);
    case LIST:
        return sizeof(ku_list_t);
    default:
        throw Exception(
            "Cannot infer the size of dataTypeID: " + dataTypeToString(dataTypeID) + ".");
    }
}

RelDirection operator!(RelDirection& direction) {
    return (FWD == direction) ? BWD : FWD;
}

std::string getRelDirectionAsString(RelDirection direction) {
    return (FWD == direction) ? "forward" : "backward";
}

} // namespace common
} // namespace kuzu
